@@ -58,6 +58,8 @@ class JacobianWrapper(object):
         dy_val        = self.rhs(y, **kwargs)
         unravelled_dy = D.reshape(dy_val, (-1,))
         jacobian_y    = D.zeros((*unravelled_y.shape, *unravelled_dy.shape), dtype=unravelled_dy.dtype)
+        if D.backend() == 'torch':
+            jacobian_y = jacobian_y.to(y.device)
         for idx,val in enumerate(unravelled_y):
             y_jac = D.copy(unravelled_y)
             dy_cur = dy
